/*
 * Commands to load factory data (FDT) partition into memory etc.
 *
 * Copyright (C) 2010 TomTom International B.V.
 * 
 ************************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 ************************************************************************
 */

#include <common.h>
#include <dsasig.h>

#ifdef CONFIG_CMD_FD

#include <command.h>
#include <part.h>
#include <libfdt.h>

#define FD_DEBUG

#ifdef	FD_DEBUG
# define PRINTF(fmt,args...)	printf (fmt ,##args)
#else
# define PRINTF(fmt,args...)
#endif /* FD_DEBUG */

#define FD_MAXSECTORS 128*1024/512

/* Load TomTom factory data partition into RAM (device-tree version) */
int do_fdload (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	char *ep;
	int dev, part = 1;
	ulong readlen, blksz, start;
	disk_partition_t info;
	block_dev_desc_t *dev_desc = NULL;
	char ulong_str[16];
	char *addr_str;

	if (argc < 3) {
		printf("Usage:\n%s\n", cmdtp->usage);
		return(1);
	}

	size_fd = 0;
	setenv("fdsize", NULL);

	if (argc == 4) {
		addr_fd = (void *) simple_strtoul (argv[3], NULL, 16);
	}
#ifdef __NON_LEGACY
	else if (argc == 3) {
		addr_str = getenv("fdaddr");
		if (addr_str != NULL) {
			addr_fd = (void *)
				simple_strtoul(addr_str, NULL, 16);
		} else {
			addr_fd = CFG_FDT_ADDR;
		}
	}
#endif

	dev = (int)simple_strtoul (argv[2], &ep, 16);
	dev_desc = get_dev (argv[1], dev);
	if (dev_desc == NULL) {
		printf("** Block device %s %d not supported\n", argv[1], dev);
		return(1);
	}

	if (*ep) {
		if (*ep != ':') {
			printf("** Invalid boot device, use `dev[:part]'\n");
			return(1);
		}
		part = (int)simple_strtoul(++ep, NULL, 16);
	}

	PRINTF("Using device %s%d, partition %d\n", argv[1], dev, part);
	
	readlen = FD_MAXSECTORS;
	blksz = dev_desc->blksz;
	start = 0;

	if (part != 0) {
		if (get_partition_info (dev_desc, part, &info)) {
			printf("** Bad partition %d\n", part);
			return(1);
		}

		if (info.size > FD_MAXSECTORS) {
			printf("** Warning - Factory Data partition size "
				"exceeds %d sectors, truncating\n",
				FD_MAXSECTORS);
		}
		else {
			readlen = info.size;
		}
		blksz = info.blksz;
		start = info.start;

		PRINTF("Loading from block device %s device %d, partition "
			"%d: Name: %.32s  Type: %.32s\n", argv[1], dev, part,
			info.name, info.type);
	} else {
		PRINTF("Loading factory data from block device %s device "
			"%d\n", argv[1], dev);
	}

	if (dev_desc->block_read(dev_desc->dev,
			start, readlen, addr_fd) != readlen) {
		printf("** Error reading factory area - start sector %ld - size %ld\n", start, readlen);
		memset(addr_fd, 0, readlen << 9);
		return(1);
	}

	/* Get real device-tree binary total size */
	if (fdt_magic(addr_fd) != FDT_MAGIC) {
		printf("** Invalid dtb magic (%x/%x)\n", fdt_magic(addr_fd), FDT_MAGIC);
		memset(addr_fd, 0, readlen << 9);
		return(1);
	}
	size_fd = fdt_totalsize(addr_fd);
	sprintf(ulong_str, "%#x", size_fd);
	setenv("fdsize", ulong_str);

	/* make sure we didn't pick up any trailing trash */
	memset(addr_fd + size_fd + DSASIG_SIZE, 0, (readlen << 9) - size_fd - DSASIG_SIZE);

	printf("Loaded device-tree at address %p of length %ld bytes\n", addr_fd, size_fd);

	return(0);
}

U_BOOT_CMD(
	fdload,	4,	0,	do_fdload,
	"fdload  - Load TomTom factory data device-tree into ram\n",
	"<interface> <dev[:part]> [<addr>]\n"
);

#endif /* CONFIG_CMD_FD */

