/*
 * verify_image.c - verify the kernel image and factory data
 *
 * 2009-11-03 Ard Biesheuvel <ard.biesheuvel@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/*
 * THEORY OF OPERATION
 * ===================
 *
 * The verify_image() function in this source file is called from the bootm handler to check
 * whether the loaded images can be allowed to boot.
 *
 * First, we check whether factory data is present at the expected location (addr_fd). If this 
 * value is NULL, we are apparently not trying to load factory data, so we ignore it. This allows
 * booting images that do not rely on factory data (i.e., a flasher zImage)
 * 
 * Next, we look at the size of the loaded factory data (size_fd). If it is zero, loading of 
 * factory data was attempted but failed. If we are booting the normal image, we abort here and wait
 * for the watchdog to kick in, which will result in the recovery kernel being selected the next time.
 * If we are already booting the recovery kernel, we spoof a factory data image containing only the
 * encrypted box key and copy it to addr_fd. This will make it available from userland in recovery
 * mode.
 * 
 * If both addr_fd and size_fd are set, we proceed to verify the factory data by calculating an HMAC
 * over the complete factory data image using the box key as secret key. The resulting hash is
 * signature checked using DSA against the factory production key first, and -if it fails- against the 
 * factory development key. If either check succeeds, we check the device-class field in the factory 
 * data to make sure it matches the factory key used (production or development). If this fails, we
 * check the factory data against the kernel key to decide whether we load it as default factory data,
 * i.e., factory data that does not contain any information specific to a single unit like serial# or
 * MAC address. If the factory data does not check out at all, we overwrite it with a spoofed version
 * containing only the encrypted box key.
 *
 * Finally, we check the DSA signature of the SHA1 hash of the kernel image using the kernel production
 * key, and retry with the kernel development key if the device class obtained from the factory data
 * allows it.
 *
 * The flow is optimized for a production device. In this case, it consists of 3 SHA1 computations (two
 * for the HMAC, one for the kernel) and two DSA signature verifications. In the worst case, it
 * consists of 3 SHA1 computations, four DSA signature verifications and one Elgamal encryption (of
 * the box key). The typical flow of a development device is 3x SHA1 and 4x DSA.
 *
 * Keys used:
 * - HMAC key 			(random 16 bytes)
 * - factory production key 	(2048 bit DSA)
 * - factory development key 	(2048 bit DSA)
 * - kernel production key 	(2048 bit DSA)
 * - kernel development key 	(2048 bit DSA)
 * - factory encryption key 	(1024 bit Elgamal)
 *
 * NOTE: in Chernobyl, 'retail' == 'production'.
 */

#include <common.h>
#include <boxkey.h>
#include <sha1.h>
#include <dsasig.h>
#include <elgamalenc.h>
#include <flipflop.h>
#include <fdt_boxkey.h>
#include <unrandom.h>
#include <devclass.h>
#include <libfdt.h>
#include <ttcrypto_ctx.h>

extern void *addr_fd;			/* location of factory data area */
extern size_t size_fd;			/* size of factory data area */
extern ulong load_size;			/* size of the loaded kernel image */

#define DCACHE_ON	dcache_enable
#define DCACHE_OFF	dcache_disable

static char *get_fdt_encrypted_boxkey(void const *fdt)
{
	int featurenode;
	char *addr;

	if (fdt_check_header(fdt)
		|| 0 > (featurenode = fdt_path_offset(fdt, "/features"))
		|| NULL == (addr = fdt_getprop(fdt, featurenode, "fdt-recovery-key", NULL)))
		return NULL;

	return addr;
}

#define LEN_DEVICE_SERIAL	(12)

static char *get_fdt_device_serial(void const *fdt)
{
	int featurenode;
	char *addr;

	if (fdt_check_header(fdt)
		|| 0 > (featurenode = fdt_path_offset(fdt, "/features"))
		|| NULL == (addr = fdt_getprop(fdt, featurenode, "device-serial", NULL)))
		return NULL;

	return addr;
}

int verify_image(void *addr)
{
	unsigned char boxkey[BOXKEY_SIZE]; 

	if (!load_size) { 
		puts ("ERROR: cannot verify image, size unknown!\n");
		return 1;
	} else  {
		unsigned char hash[20];
		int dev_class = DEVCLASS_PRODUCTION;

		DCACHE_ON();

		sha1_csum((void*)addr, load_size - DSASIG_SIZE, hash);

		/*
		 * addr_fd is set after we have attempted to load a device tree (but may have failed)
		 * size_fd is set after a syntactically correct device tree has been loaded
		 */
		if (addr_fd) {
			unsigned char hmac[20];

			printf("Verifying device tree...\n");

			get_boxkey(boxkey);

			if (size_fd)
				sha1_hmac(boxkey, BOXKEY_SIZE, addr_fd, size_fd, hmac);

			/* take care not to touch the factory data until after it has been verified,
			   but make sure it is consistent with the key used */

			if (!size_fd ||
				((dsa_verify_hash(hmac, addr_fd + size_fd, FACTORY_PROD_KEY)
					|| DEVCLASS_PRODUCTION != get_dev_class(addr_fd))
				&& (dsa_verify_hash(hmac, addr_fd + size_fd, FACTORY_DEV_KEY)
					|| DEVCLASS_DEVELOPMENT != (dev_class = get_dev_class(addr_fd)))))
			{
				char *boxkey_field;
				/*
				 * The flipflop will be set if we are about to boot a regular image,
				 * and it will be cleared if we are not => this is opposite to how
				 * it is used in other places.
				 */
				if (flipflop_get()) {
					printf("Normal boot aborted, factory data invalid or missing!\n");
#ifdef __NON_LEGACY
					DCACHE_OFF();
					return 1;
#else
					hang();
#endif
				}

				if (size_fd)
					sha1_csum(addr_fd, size_fd, hmac);

				/* we allow generic factory data only if it is signed by TomTom */
				if (!size_fd
					|| dsa_verify_hash(hmac, addr_fd + size_fd, KERNEL_PROD_KEY)
					|| !(boxkey_field = get_fdt_encrypted_boxkey(addr_fd)))
				{
					printf("Continuing backup boot with built-in device tree\n");

					size_fd = sizeof(pbk_fdt);
					memcpy(addr_fd, &pbk_fdt, size_fd);
					boxkey_field = ((struct _pbk_fdt*)addr_fd)->key;
				}

				elgamal_encrypt(boxkey_field, boxkey, BOXKEY_SIZE);

			} else {
				/* we have valid per-device factory data, so put the shared secret in memory  */
				char *serial = get_fdt_device_serial(addr_fd);

				if (serial) {
					struct ttcrypto_ctx *ctx;

					TTCRYPTO_INIT(ctx,addr_fd,size_fd);

					sha1_hmac(boxkey, BOXKEY_SIZE, serial, LEN_DEVICE_SERIAL,
						ctx->shared_secret);
				}
			}
			memset(&boxkey, 0, BOXKEY_SIZE);
		}

		/*
		 * verify the image using the signature that is in the last DSASIG_SIZE
		 * bytes of the loaded image
		 */

		printf("Verifying executable image signature ...\n");

		if (dsa_verify_hash(hash, (void*)addr + (load_size - DSASIG_SIZE), KERNEL_PROD_KEY)
			&& (dev_class != DEVCLASS_DEVELOPMENT 
				|| dsa_verify_hash(hash, (void*)addr + (load_size - DSASIG_SIZE), KERNEL_DEV_KEY)))
		{
			puts ("ERROR: invalid image, signature check failed!\n");
			memset((void*)addr, 0, load_size);
			DCACHE_OFF();
			return 1;
		}
	}

	DCACHE_OFF();
	UNRANDOM_FINI();

	return 0;
}

