/*
 * u-boot titan MMC/SD card driver
 * (C) Copyright SiRF Technology, Inc
 *
 *
 * based on u-boot pxa MMC driver and linux/drivers/mmc/core.c
 * (C) 2005-2005 Thomas Kleffel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <common.h>
#include <mmc.h>
#include <asm/io.h>
#include <asm/errno.h>
#include <part.h>
#include <fat.h>

#include "titan.h"
#include <asm/arch/sirfsoc.h>
#include <asm/arch/sirfsoc_clocks.h>

#ifdef CONFIG_MMC

//#define TITAN_MMC_DEBUG 1
#ifdef TITAN_MMC_DEBUG
#define mmc_debug(fmt,args...)	printf(fmt ,##args)
#else
#define mmc_debug(fmt,args...)
#endif 
#define TITAN_MMC_ERROR_LOG 1
#ifdef TITAN_MMC_ERROR_LOG
#define mmc_err(fmt,args...)	printf(fmt ,##args)
#else
#define mmc_err(fmt,args...)
#endif 

#ifdef CONFIG_SIRFSOC_FPGA
#define MAX_CARD_CLOCK_RATE 10000000
#else
#define MAX_CARD_CLOCK_RATE 20000000
#endif



#define MMC_RSP_PRESENT	(1 << 0)
#define MMC_RSP_136	(1 << 1)		/* 136 bit response */
#define MMC_RSP_CRC	(1 << 2)		/* expect valid crc */
#define MMC_RSP_BUSY	(1 << 3)		/* card may send busy */
#define MMC_RSP_OPCODE	(1 << 4)		/* response contains opcode */
#define MMC_RSP_DATA	(1 << 5)		/* data follows the response*/

#define MMC_RSP_NONE	(0)
#define MMC_RSP_R1	(MMC_RSP_PRESENT|MMC_RSP_CRC|MMC_RSP_OPCODE)
#define MMC_RSP_R1B	(MMC_RSP_PRESENT|MMC_RSP_CRC|MMC_RSP_OPCODE|MMC_RSP_BUSY)
#define MMC_RSP_R2	(MMC_RSP_PRESENT|MMC_RSP_136|MMC_RSP_CRC)
#define MMC_RSP_R3	(MMC_RSP_PRESENT)
#define MMC_RSP_R6	(MMC_RSP_PRESENT|MMC_RSP_CRC|MMC_RSP_OPCODE)
#define MMC_RSP_R7	(MMC_RSP_PRESENT|MMC_RSP_CRC|MMC_RSP_OPCODE)

/*
 * FIXME needs to read cid and csd info to determine block size
 * and other parameters
 */
static uchar mmc_buf[MMC_BLOCK_SIZE];
struct mmc_csd csd;
uint raw_csd[4];
struct mmc_cid cid;
uint raw_cid[4];
static uint rca = 0;
uint raw_scr[2];
struct sd_scr scr;


static uint mmc_ready = 0;
static uint highspeed = 0;
static uint highcapacity = 0;
static uint ocr_avail = 0;
static uint ocr_selected = 0;

static const unsigned int tran_exp[] = {
	10000,		100000,		1000000,	10000000,
	0,		0,		0,		0
};

static const unsigned char tran_mant[] = {
	0,	10,	12,	13,	15,	20,	25,	30,
	35,	40,	45,	50,	55,	60,	70,	80,
};

static const unsigned int tacc_exp[] = {
	1,	10,	100,	1000,	10000,	100000,	1000000, 10000000,
};

static const unsigned int tacc_mant[] = {
	0,	10,	12,	13,	15,	20,	25,	30,
	35,	40,	45,	50,	55,	60,	70,	80,
};

static block_dev_desc_t mmc_dev;

block_dev_desc_t* mmc_get_dev(int dev)
{
	return( (block_dev_desc_t*) &mmc_dev);
}

/*
 * Given the decoded CSD structure, decode the raw CID to our CID structure.
 */
static int mmc_decode_cid_mmc(void)
{
	u32 *resp = raw_cid;

	/*
	 * The selection of the format here is based upon published
	 * specs from sandisk and from what people have reported.
	 */
	switch (csd.mmca_vsn) {
		case 0: /* MMC v1.0 - v1.2 */
		case 1: /* MMC v1.4 */
			cid.manfid	= UNSTUFF_BITS(resp, 104, 24);
			cid.prod_name[0]	= UNSTUFF_BITS(resp, 96, 8);
			cid.prod_name[1]	= UNSTUFF_BITS(resp, 88, 8);
			cid.prod_name[2]	= UNSTUFF_BITS(resp, 80, 8);
			cid.prod_name[3]	= UNSTUFF_BITS(resp, 72, 8);
			cid.prod_name[4]	= UNSTUFF_BITS(resp, 64, 8);
			cid.prod_name[5]	= UNSTUFF_BITS(resp, 56, 8);
			cid.prod_name[6]	= UNSTUFF_BITS(resp, 48, 8);
			cid.hwrev		= UNSTUFF_BITS(resp, 44, 4);
			cid.fwrev		= UNSTUFF_BITS(resp, 40, 4);
			cid.serial	= UNSTUFF_BITS(resp, 16, 24);
			cid.month		= UNSTUFF_BITS(resp, 12, 4);
			cid.year		= UNSTUFF_BITS(resp, 8, 4) + 1997;
			break;

		case 2: /* MMC v2.0 - v2.2 */
		case 3: /* MMC v3.1 - v3.3 */
		case 4: /* MMC v4 */
			cid.manfid	= UNSTUFF_BITS(resp, 120, 8);
			cid.oemid		= UNSTUFF_BITS(resp, 104, 16);
			cid.prod_name[0]	= UNSTUFF_BITS(resp, 96, 8);
			cid.prod_name[1]	= UNSTUFF_BITS(resp, 88, 8);
			cid.prod_name[2]	= UNSTUFF_BITS(resp, 80, 8);
			cid.prod_name[3]	= UNSTUFF_BITS(resp, 72, 8);
			cid.prod_name[4]	= UNSTUFF_BITS(resp, 64, 8);
			cid.prod_name[5]	= UNSTUFF_BITS(resp, 56, 8);
			cid.serial	= UNSTUFF_BITS(resp, 16, 32);
			cid.month		= UNSTUFF_BITS(resp, 12, 4);
			cid.year		= UNSTUFF_BITS(resp, 8, 4) + 1997;
			break;

		default:
			mmc_err("TITANMMC: card has unknown MMCA version %d\n",
					csd.mmca_vsn);
			return -EINVAL;
	}

	return 0;
}

/*
 * Given a 128-bit response, decode to our card CSD structure.
 */
static int mmc_decode_csd_mmc(void)
{
	unsigned int e, m, csd_struct;
	u32 *resp = raw_csd;

	/*
	 * We only understand CSD structure v1.1 and v1.2.
	 * v1.2 has extra information in bits 15, 11 and 10.
	 */
	csd_struct = UNSTUFF_BITS(resp, 126, 2);
	if (csd_struct != 1 && csd_struct != 2) {
		mmc_err("TITANMMC: unrecognised CSD structure version %d\n",
				csd_struct);
		return -EINVAL;
	}

	csd.mmca_vsn	 = UNSTUFF_BITS(resp, 122, 4);
	m = UNSTUFF_BITS(resp, 115, 4);
	e = UNSTUFF_BITS(resp, 112, 3);
	csd.tacc_ns	 = (tacc_exp[e] * tacc_mant[m] + 9) / 10;
	csd.tacc_clks	 = UNSTUFF_BITS(resp, 104, 8) * 100;

	m = UNSTUFF_BITS(resp, 99, 4);
	e = UNSTUFF_BITS(resp, 96, 3);
	csd.max_dtr	  = tran_exp[e] * tran_mant[m];
	csd.cmdclass	  = UNSTUFF_BITS(resp, 84, 12);

	e = UNSTUFF_BITS(resp, 47, 3);
	m = UNSTUFF_BITS(resp, 62, 12);
	csd.capacity	  = (1 + m) << (e + 2);

	csd.read_blkbits = UNSTUFF_BITS(resp, 80, 4);
	csd.read_partial = UNSTUFF_BITS(resp, 79, 1);
	csd.write_misalign = UNSTUFF_BITS(resp, 78, 1);
	csd.read_misalign = UNSTUFF_BITS(resp, 77, 1);
	csd.r2w_factor = UNSTUFF_BITS(resp, 26, 3);
	csd.write_blkbits = UNSTUFF_BITS(resp, 22, 4);
	csd.write_partial = UNSTUFF_BITS(resp, 21, 1);

	return 0;
}

static void mmc_decode_cid_sd(void)
{
	u32 *resp = raw_cid;

	memset(&cid, 0, sizeof(struct mmc_cid));

	/*
	 * SD doesn't currently have a version field so we will
	 * have to assume we can parse this.
	 */
	cid.manfid		= UNSTUFF_BITS(resp, 120, 8);
	cid.oemid			= UNSTUFF_BITS(resp, 104, 16);
	cid.prod_name[0]		= UNSTUFF_BITS(resp, 96, 8);
	cid.prod_name[1]		= UNSTUFF_BITS(resp, 88, 8);
	cid.prod_name[2]		= UNSTUFF_BITS(resp, 80, 8);
	cid.prod_name[3]		= UNSTUFF_BITS(resp, 72, 8);
	cid.prod_name[4]		= UNSTUFF_BITS(resp, 64, 8);
	cid.hwrev			= UNSTUFF_BITS(resp, 60, 4);
	cid.fwrev			= UNSTUFF_BITS(resp, 56, 4);
	cid.serial		= UNSTUFF_BITS(resp, 24, 32);
	cid.year			= UNSTUFF_BITS(resp, 12, 8);
	cid.month			= UNSTUFF_BITS(resp, 8, 4);

	cid.year += 2000; /* SD cards year offset */
}

/*
 * Given a 128-bit response, decode to our card CSD structure.
 */
static int mmc_decode_csd_sd(void)
{
	unsigned int e, m, csd_struct;
	u32 *resp = raw_csd;

	csd_struct = UNSTUFF_BITS(resp, 126, 2);

	switch (csd_struct) {
		case 0:
			m = UNSTUFF_BITS(resp, 115, 4);
			e = UNSTUFF_BITS(resp, 112, 3);
			csd.tacc_ns	 = (tacc_exp[e] * tacc_mant[m] + 9) / 10;
			csd.tacc_clks	 = UNSTUFF_BITS(resp, 104, 8) * 100;

			m = UNSTUFF_BITS(resp, 99, 4);
			e = UNSTUFF_BITS(resp, 96, 3);
			csd.max_dtr	  = tran_exp[e] * tran_mant[m];
			csd.cmdclass	  = UNSTUFF_BITS(resp, 84, 12);

			e = UNSTUFF_BITS(resp, 47, 3);
			m = UNSTUFF_BITS(resp, 62, 12);
			csd.capacity	  = (1 + m) << (e + 2);

			csd.read_blkbits = UNSTUFF_BITS(resp, 80, 4);
			csd.read_partial = UNSTUFF_BITS(resp, 79, 1);
			csd.write_misalign = UNSTUFF_BITS(resp, 78, 1);
			csd.read_misalign = UNSTUFF_BITS(resp, 77, 1);
			csd.r2w_factor = UNSTUFF_BITS(resp, 26, 3);
			csd.write_blkbits = UNSTUFF_BITS(resp, 22, 4);
			csd.write_partial = UNSTUFF_BITS(resp, 21, 1);
			break;
		case 1:
			/*
			 * This is a block-addressed SDHC card. Most
			 * interesting fields are unused and have fixed
			 * values. To avoid getting tripped by buggy cards,
			 * we assume those fixed values ourselves.
			 */
			highcapacity = 1;

			csd.tacc_ns	 = 0; /* Unused */
			csd.tacc_clks	 = 0; /* Unused */

			m = UNSTUFF_BITS(resp, 99, 4);
			e = UNSTUFF_BITS(resp, 96, 3);
			csd.max_dtr	  = tran_exp[e] * tran_mant[m];
			csd.cmdclass	  = UNSTUFF_BITS(resp, 84, 12);

			m = UNSTUFF_BITS(resp, 48, 22);
			csd.capacity     = (1 + m) << 10;

			csd.read_blkbits = 9;
			csd.read_partial = 0;
			csd.write_misalign = 0;
			csd.read_misalign = 0;
			csd.r2w_factor = 4; /* Unused */
			csd.write_blkbits = 9;
			csd.write_partial = 0;
			break;
		default:
			mmc_err("TITANMMC unrecognised CSD structure version %d\n",
					csd_struct);
			return -EINVAL;
	}

	return 0;
}
/* Decodes the bus width field of the scr register */
static int mmc_decode_scr(void)
{
	unsigned int scr_struct;
	u32 resp[4];

	resp[3] = raw_scr[1];
	resp[2] = raw_scr[0];

	scr_struct = UNSTUFF_BITS(resp, 60, 4);
	if (scr_struct != 0) {
		printf("TITANMMC: unrecognised SCR structure version %d\n",
				scr_struct);
		return -EINVAL;
	}

	scr.sda_vsn = UNSTUFF_BITS(resp, 56, 4);
	scr.bus_widths = UNSTUFF_BITS(resp, 48, 4);

	return 0;
}

static void print_cid(void)
{
	printf("Manufacturer ID = 0x%x\n", cid.manfid);
	printf("HW/FW Revision 	= 0x%x 0x%x\n", cid.hwrev, cid.fwrev);
	printf("Product Name 	= %s\n", cid.prod_name);
	printf("Serial Number 	= 0x%x\n", cid.serial);
	printf("OEM id 		= 0x%x\n", cid.oemid);
	printf("Month 		= %d\n", cid.month);
	printf("Year 		= %d\n", cid.year);
}

/* Controller reset; Should be called after any error  */
static void mmc_reset(u8 mask)
{
	unsigned long timeout;

	writeb(mask, TITANMMC_PHY_BASE + TITANMMC_SOFTWARE_RESET);

	/* Wait max 100 ms */
	timeout = 100;

	/* hw clears the bit when it's done */
	while (readb(TITANMMC_PHY_BASE + TITANMMC_SOFTWARE_RESET) & mask) {
		if (timeout == 0) {
			mmc_err("TITANMMC: Reset never completed.\n");
			return;
		}
		timeout--;
		udelay(1000);
	}
	mmc_debug("TITANMMC: Reset Complete\n");
	return;
}

/* Waits for inhibit bits in the present state register to 
 * get clear */
static int mmc_wait_inhibit_clear( int mask)
{
	int timeout;
	timeout = 100;

	while (readl(TITANMMC_PHY_BASE + TITANMMC_PRESENT_STATE) & mask) {
		if (timeout == 0) {
			mmc_err("TITANMMC: Controller"
					" never released inhibit bit(s).\n");
			return -1;
		}
		timeout--;
		udelay(1000);
	}
	return 0;
}
/* Waits for the interrupts to set, clears and returns the interrupt bits
 * that were set 
 * If not set in 5 mins returns 0*/
static inline int mmc_wait_interrupt(int mask)
{
	int timeout;
	uint status;
	timeout = 100000000;
	while((status = readl(TITANMMC_PHY_BASE + TITANMMC_INT_STATUS)& mask) == 0){
		if (timeout == 0) {
			mmc_err("TITANMMC: interrupt timed out waiting for mask 0x%x status is 0x%x\n", mask, status);
			return 0;
		}
		timeout--;
#ifdef CONFIG_T32MON
		udelay(10);
#endif
	}
	/* clear the interrupts */
	writel(status, TITANMMC_PHY_BASE + TITANMMC_INT_STATUS);

	mmc_debug("Got interrupt 0x%x\n", status);
	return status;
}

static uint *mmc_cmd(ushort cmd, ulong arg, uint response_flags)
{
	static uint resp[5];
	uint intmask;
	uint flags, i;
	uint inhibit_mask = 0;
	memset(resp, 0, sizeof(resp));

	mmc_debug("mmc_cmd CMD%d arg=0x%08x flags=%x\n", cmd, arg, response_flags);

	/* check for inhibit */
	inhibit_mask = TITANMMC_CMD_INHIBIT;
	if(response_flags & MMC_RSP_DATA)
		inhibit_mask |= TITANMMC_DATA_INHIBIT;
	if(mmc_wait_inhibit_clear(inhibit_mask))
		return NULL;
	/* write the argument in the argument register */
	writel(arg, TITANMMC_PHY_BASE + TITANMMC_ARGUMENT);
	/* set the flags */
	if (!(response_flags & MMC_RSP_PRESENT))
		flags = TITANMMC_CMD_RESP_NONE;
	else if (response_flags & MMC_RSP_136)
		flags = TITANMMC_CMD_RESP_LONG;
	else if (response_flags & MMC_RSP_BUSY)
		flags = TITANMMC_CMD_RESP_SHORT_BUSY;
	else
		flags = TITANMMC_CMD_RESP_SHORT;

	if (response_flags & MMC_RSP_CRC)
		flags |= TITANMMC_CMD_CRC;
	if (response_flags & MMC_RSP_OPCODE)
		flags |= TITANMMC_CMD_INDEX;
	if (response_flags & MMC_RSP_DATA)
		flags |= TITANMMC_CMD_DATA;

	/* make command using the command opcode and the flags
	 * and write to the command register */
	writew(TITANMMC_MAKE_CMD(cmd, flags), TITANMMC_PHY_BASE + TITANMMC_COMMAND);

	/* wait for response interrupt/error interrupt/bad crc */
	intmask = mmc_wait_interrupt(TITANMMC_INT_CMD_MASK);
	if(!intmask)
		return NULL;
	/* check for error */
	if (intmask & TITANMMC_INT_RESPONSE){
		if (response_flags & MMC_RSP_PRESENT) {
			if (response_flags & MMC_RSP_136) {
				/* CRC is stripped so we need to do some shifting. */
				for (i = 0;i < 4;i++) {
					resp[i] = readl(TITANMMC_PHY_BASE +
							TITANMMC_RESPONSE + (3-i)*4) << 8;
					if (i != 3)
						resp[i] |=
							readb(TITANMMC_PHY_BASE +
									TITANMMC_RESPONSE + (3-i)*4-1);
				}
			} else {
				resp[0] = readl(TITANMMC_PHY_BASE + TITANMMC_RESPONSE);
			}
		}
	}
	else if( (intmask & (TITANMMC_INT_TIMEOUT | 
					TITANMMC_INT_CRC | 
					TITANMMC_INT_END_BIT | 
					TITANMMC_INT_INDEX)))
	{
		mmc_debug("TITANMMC: Got error interrupt 0x%x\n", intmask);
		mmc_reset(TITANMMC_RESET_CMD | TITANMMC_RESET_DATA);
		return NULL;
	}

	if (!mmc_ready)
		udelay(5000);
	return resp;
}
static int mmc_block_read(uchar *dst, ulong src, ulong len)
{
	uint *dst_u32 = (uint *)dst;
	uint *resp;
	uint intmask;
	if (len == 0)
		return 0;

	mmc_debug("mmc_block_rd dst %lx src %lx len %d\n", 
			(ulong)dst, src, len);

	writeb(0xe, TITANMMC_PHY_BASE + TITANMMC_TIMEOUT_CONTROL);
	writew(TITANMMC_MAKE_BLKSZ(7, len), TITANMMC_PHY_BASE + TITANMMC_BLOCK_SIZE);
	writew(1, TITANMMC_PHY_BASE +TITANMMC_BLOCK_COUNT);

	/* setup data */
	writew(TITANMMC_TRNS_READ, TITANMMC_PHY_BASE + TITANMMC_TRANSFER_MODE);
	/* the byte address or the block address */
	if(highcapacity)
		src >>= 9;
	/* send read command */
	resp = mmc_cmd(MMC_CMD_READ_BLOCK, src, MMC_RSP_R1 | MMC_RSP_DATA);
	if(!resp)
		return -1;

	intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK);
	if(!intmask)
		return -1;
	if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
				TITANMMC_INT_DATA_CRC|
				TITANMMC_INT_DATA_END_BIT)) {
		mmc_err("mmc_block_read: err intmask%08x\n", intmask);
		return -EIO;
	}

	while (len>0) {
		*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
		if (len >= 4)
			len -= 4;
		else {
			len = 0;
			break;
		}
	}

	mmc_debug("waiting for data complete interrupt\n");
	intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK );
	if(!intmask)
		return -EIO;

	if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
				TITANMMC_INT_DATA_CRC|
				TITANMMC_INT_DATA_END_BIT)) {
		mmc_debug("mmc_block_read: err intmask %08x\n", intmask);
		return -EIO;
	}

	return 0;
}

int mmc_multi_write(ulong dst, uchar * src, int len)
{
	uint *src_u32 = (uint *)src;
	uint *resp;
	uint intmask;
	uint nr_sectors = 0;
	uint sectors;
	uint i,j;
	if (len == 0)
		return 0;

	mmc_debug("mmc_block_wr src %lx dst %lx len %d\n", 
			(ulong)src, dst, len);

	nr_sectors = len>>9;
	while(nr_sectors > 0)
	{
		mmc_debug("nrsectors 0x%x dst 0x%x \n", nr_sectors, dst);
		if(nr_sectors > 0xFFFF)
			sectors = 0xFFFF;
		else
			sectors = nr_sectors;
		writeb(0xe, TITANMMC_PHY_BASE + TITANMMC_TIMEOUT_CONTROL);
		writel(TITANMMC_MAKE_BLKSZ(7, MMC_BLOCK_SIZE) | (sectors<< 16), TITANMMC_PHY_BASE + TITANMMC_BLOCK_SIZE);

		/* setup data */
		writew(TITANMMC_TRNS_BLK_CNT_EN |TITANMMC_TRNS_MULTI | TITANMMC_TRNS_ACMD12,
				TITANMMC_PHY_BASE + TITANMMC_TRANSFER_MODE);
		/* the byte address or the block address */
		if(highcapacity)
			dst >>= 9;
		/* send read command */

		resp = mmc_cmd(MMC_CMD_WR_BLK_MULTI, dst, MMC_RSP_R1 | MMC_RSP_DATA);
		if(!resp)
		{
			return -1;
		}

		for(i=0; i<sectors;i++)
		{
			mmc_debug("Waiting for space avail interrupt\n");
			intmask =  mmc_wait_interrupt(TITANMMC_INT_SPACE_AVAIL);         

			if(!intmask){
				mmc_err("Time out waiting for space avail interrupt\n");
				return -EIO;
			}
			if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
						TITANMMC_INT_DATA_CRC|
						TITANMMC_INT_DATA_END_BIT)) {
				mmc_err("mmc_multi_write: err intmask%08x\n", intmask);
				return -EIO;
			}

			for(j=MMC_BLOCK_SIZE>>5;j>0;j--) {
				//Loop unrolled
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
			}
		}
		mmc_debug("waiting for data complete interrupt\n");
		intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK);
		if(!intmask){
			return -EIO;
		}
		if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
					TITANMMC_INT_DATA_CRC|
					TITANMMC_INT_DATA_END_BIT)) {
			mmc_debug("mmc_multi_write: err intmask %08x\n", intmask);
			return -EIO;
		}
		dst += sectors*MMC_BLOCK_SIZE;
		nr_sectors -= sectors;
	}
	return 0;
}


int mmc_multi_read(uchar *dst, ulong src, int len)
{
	uint *dst_u32 = (uint *)dst;
	uint *resp;
	uint intmask;
	uint nr_sectors = 0;
	uint sectors;
	uint i,j;
	if (len == 0)
		return 0;

	mmc_debug("mmc_block_rd dst %lx src %lx len %d\n", 
			(ulong)dst, src, len);

	/* send command for setting the block len */

	nr_sectors = len>>9;
	while(nr_sectors > 0)
	{
		mmc_debug("nrsectors 0x%x src 0x%x \n", nr_sectors, src);
		if(nr_sectors > 0xFFFF)
			sectors = 0xFFFF;
		else
			sectors = nr_sectors;
		writeb(0xe, TITANMMC_PHY_BASE + TITANMMC_TIMEOUT_CONTROL);
		writel(TITANMMC_MAKE_BLKSZ(7, MMC_BLOCK_SIZE) | (sectors<< 16), TITANMMC_PHY_BASE + TITANMMC_BLOCK_SIZE);

		/* setup data */
		writew(TITANMMC_TRNS_READ | TITANMMC_TRNS_BLK_CNT_EN |TITANMMC_TRNS_MULTI | TITANMMC_TRNS_ACMD12,
				TITANMMC_PHY_BASE + TITANMMC_TRANSFER_MODE);
		/* the byte address or the block address */
		if(highcapacity)
			src >>= 9;
		/* send read command */

		resp = mmc_cmd(MMC_CMD_RD_BLK_MULTI, src, MMC_RSP_R1 | MMC_RSP_DATA);
		if(!resp)
		{
			return -1;
		}

		for(i=0; i<sectors;i++)
		{
			mmc_debug("Waiting for data avail interrupt\n");
			intmask =  mmc_wait_interrupt(TITANMMC_INT_DATA_AVAIL);         

			if(!intmask){
				return -EIO;
			}
			if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
						TITANMMC_INT_DATA_CRC|
						TITANMMC_INT_DATA_END_BIT)) {
				mmc_err("mmc_multi_read: err intmask%08x\n", intmask);
				return -EIO;
			}

			for(j=MMC_BLOCK_SIZE>>5;j>0;j--) {
				//Loop unrolled
				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);

				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
				*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);
			}
		}
		mmc_debug("waiting for data complete interrupt\n");
		intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK);
		if(!intmask){
			return -EIO;
		}
		if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
					TITANMMC_INT_DATA_CRC|
					TITANMMC_INT_DATA_END_BIT)) {
			mmc_debug("mmc_multi_read: err intmask %08x\n", intmask);
			return -EIO;
		}
		src += sectors*MMC_BLOCK_SIZE;
		nr_sectors -= sectors;
	}
	return 0;
}

static ulong mmc_bread (int dev_num, ulong blknr, ulong blkcnt, ulong* dst)
{
	ulong src, len;
	
	src = blknr*MMC_BLOCK_SIZE;
	len = blkcnt*MMC_BLOCK_SIZE;	

	if (blkcnt != 0)
		mmc_multi_read((uchar*) dst, src, len);

	return blkcnt;
}

int mmc_block_write(ulong dst, uchar *src, int len)
{
	uint *src_u32 = (uint *)src;
	uint *resp;
	uint intmask;
	if (len == 0)
		return 0;

	mmc_debug("mmc_block_wr dst %lx src %lx len %d\n", 
			dst,(ulong) src, len);

	writeb(0xE, TITANMMC_PHY_BASE + TITANMMC_TIMEOUT_CONTROL);
	writew(TITANMMC_MAKE_BLKSZ(7, len), TITANMMC_PHY_BASE + TITANMMC_BLOCK_SIZE);
	writew(1, TITANMMC_PHY_BASE +TITANMMC_BLOCK_COUNT);

	/* setup data */
	writew(0, TITANMMC_PHY_BASE + TITANMMC_TRANSFER_MODE);
	/* the byte address or the block address */
	if(highcapacity)
		dst >>= 9;
	/* send write command */
	resp = mmc_cmd(MMC_CMD_WRITE_BLOCK, dst, MMC_RSP_R1 | MMC_RSP_DATA);
	if(!resp)
		return -1;

	intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK);
	if(!intmask)
		return -1;

	if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
				TITANMMC_INT_DATA_CRC|
				TITANMMC_INT_DATA_END_BIT)) {
		mmc_debug("mmc_block_write: err intmask %08x\n", intmask);
		return -EIO;
	}

	while (len>0) {
		writel(*(src_u32++), TITANMMC_PHY_BASE + TITANMMC_BUFFER);
		if (len >= 4)
			len -= 4;
		else {
			len = 0;
			break;
		}
	}

	mmc_debug("waiting for data complete interrupt\n");
	intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK );
	if(!intmask)
		return -EIO;
	if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
				TITANMMC_INT_DATA_CRC|
				TITANMMC_INT_DATA_END_BIT)) {
		mmc_debug("mmc_block_write: err intmask %08x\n", intmask);
		return -EIO;
	}

	return 0;
}

int mmc_read(ulong src, uchar *dst, int size)
{
	ulong mmc_block_size;

	if (size == 0)
		return 0;

	if (!mmc_ready) {
		mmc_debug("Please initialize the MMC first\n");
		return -1;
	}

	mmc_block_size = MMC_BLOCK_SIZE;

	src -= CFG_MMC_BASE;

	if(mmc_multi_read(dst, src, size) < 0)
		return -1;
	else
		return 0;
}

int mmc_write(uchar *src, ulong dst, int size)
{
	ulong end, part_start, part_end, part_len, aligned_start, aligned_end;
	ulong mmc_block_size, mmc_block_address;

	if (size == 0)
		return 0;

	if (!mmc_ready) {
		mmc_debug("Please initialize the MMC first\n");
		return -1;
	}

	mmc_block_size = MMC_BLOCK_SIZE;
	mmc_block_address = ~(mmc_block_size - 1);

	dst -= CFG_MMC_BASE;
	end = dst + size;
	part_start = ~mmc_block_address & dst;
	part_end = ~mmc_block_address & end;
	aligned_start = mmc_block_address & dst;
	aligned_end = mmc_block_address & end;

	/* all block aligned accesses */
	mmc_debug("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
			src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
	if (part_start) {
		part_len = mmc_block_size - part_start;
		mmc_debug("ps src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
				(ulong)src, dst, end, part_start, part_end, aligned_start, aligned_end);
		if ((mmc_block_read(mmc_buf, aligned_start, mmc_block_size)) < 0)
			return -1;

		memcpy(mmc_buf+part_start, src, part_len);
		if ((mmc_block_write(aligned_start, mmc_buf, mmc_block_size)) < 0)
			return -1;

		dst += part_len;
		src += part_len;
	}
	mmc_debug("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
			src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
	for (; dst < aligned_end; src += mmc_block_size, dst += mmc_block_size) {
		mmc_debug("al src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
				src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
		if ((mmc_block_write(dst, (uchar *)src, mmc_block_size)) < 0)
			return -1;

	}
	mmc_debug("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
			src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
	if (part_end && dst < end) {
		mmc_debug("pe src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
				src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
		if ((mmc_block_read(mmc_buf, aligned_end, mmc_block_size)) < 0)
			return -1;

		memcpy(mmc_buf, src, part_end);
		if ((mmc_block_write(aligned_end, mmc_buf, mmc_block_size)) < 0)
			return -1;
	}
	return 0;
}

static void mmc_enable_interrupt(void)
{
	uint intmask;

	intmask = TITANMMC_INT_BUS_POWER | TITANMMC_INT_DATA_END_BIT |
		TITANMMC_INT_DATA_CRC | TITANMMC_INT_DATA_TIMEOUT 
		| TITANMMC_INT_INDEX |	TITANMMC_INT_END_BIT | 
		TITANMMC_INT_CRC | TITANMMC_INT_TIMEOUT |
		TITANMMC_INT_DATA_AVAIL | TITANMMC_INT_SPACE_AVAIL |
		TITANMMC_INT_DMA_END | TITANMMC_INT_DATA_END | 
		TITANMMC_INT_RESPONSE;

	writel(intmask, TITANMMC_PHY_BASE + TITANMMC_INT_ENABLE);
}

static void mmc_set_power(uint vdd)
{
	uint power_reg = TITANMMC_POWER_ON;
	power_reg |= vdd;
	writeb(power_reg, TITANMMC_PHY_BASE + TITANMMC_POWER_CONTROL);
	return;
}

/* sets the divider value such
 * that the frequency is closest to the
 * desired clk */
static void mmc_set_clock(uint freq)
{
	uint max_clk;
	int div;
	int clkreg_value = 0;
	unsigned long timeout;
	const struct sirfsoc_clocks *clk;

	clk = sirfsoc_clocks_get();

	max_clk = clk->sd;
	mmc_debug("SD Clock = %d\n", max_clk);
	writew(0, TITANMMC_PHY_BASE + TITANMMC_CLOCK_CONTROL);

	for (div = 1;div < 256;div *= 2) {
		if ((max_clk / div) <= freq)
			break;
	}

	/* divide the divisor by 2 to program to the controller */
	div >>= 1;

	clkreg_value = div << TITANMMC_DIVIDER_SHIFT;
	clkreg_value |= TITANMMC_CLOCK_INT_EN;

	writew(clkreg_value, TITANMMC_PHY_BASE + TITANMMC_CLOCK_CONTROL);

	/* Wait max 10 ms */
	timeout = 100;

	while (!((clkreg_value = readw(TITANMMC_PHY_BASE + TITANMMC_CLOCK_CONTROL))
				& TITANMMC_CLOCK_INT_STABLE)) {
		if (timeout == 0) {
			mmc_debug("TITANMMC: Internal clock never "
					"stabilised.\n");
			return;
		}
		timeout--;
		udelay(100);
	}

	clkreg_value |= TITANMMC_CLOCK_CARD_EN;
	writew(clkreg_value, TITANMMC_PHY_BASE + TITANMMC_CLOCK_CONTROL);

	return;
}

static void mmc_set_buswidth(uint busw)
{
	uint ctrl = 0;
	ctrl = readb(TITANMMC_PHY_BASE + TITANMMC_HOST_CONTROL);

	/*Check for bus width and speed and write it into host controller register.*/

	if (busw == MMC_BUS_WIDTH_4)
	{
		mmc_debug("setting 4 bit bus \n");
		ctrl |= TITANMMC_CTRL_4BITBUS;
		ctrl &= ~TITANMMC_CTRL_8BITBUS;
	}
	else if(busw == MMC_BUS_WIDTH_8)
	{
		mmc_debug("setting 8 bit bus \n");
		ctrl |= TITANMMC_CTRL_8BITBUS;
		ctrl &= ~TITANMMC_CTRL_4BITBUS;
	}	
	else
	{
		mmc_debug("setting 1 bit bus \n");
		ctrl &= ~(TITANMMC_CTRL_4BITBUS | TITANMMC_CTRL_8BITBUS);
	}
	writeb(ctrl, TITANMMC_PHY_BASE + TITANMMC_HOST_CONTROL);
	return;
}

static void mmc_power_off(void)
{
	writew(0, TITANMMC_PHY_BASE + TITANMMC_CLOCK_CONTROL);
	writeb(0, TITANMMC_PHY_BASE + TITANMMC_POWER_CONTROL);
	return;
}

static void mmc_power_up(void)
{
	uint caps = 0;
	uint vdd = 0;
	caps	= readl(TITANMMC_PHY_BASE + TITANMMC_CAPABILITIES);
	/* select the highest voltage supported */
	if (caps & TITANMMC_CAN_VDD_330)
		vdd = TITANMMC_POWER_330;
	else if (caps & TITANMMC_CAN_VDD_300)
		vdd = TITANMMC_POWER_300;
	else if (caps & TITANMMC_CAN_VDD_180)
		vdd = TITANMMC_POWER_180;

	mmc_set_power(vdd);

	/* select the proper init clock (400KHz)*/
	mmc_set_clock(400000); 
	/* select the 1 bit bus */
	mmc_set_buswidth(1);

	return;
}

static void mmc_set_avail_ocr(void)
{
	uint caps = 0;
	caps	= readl(TITANMMC_PHY_BASE + TITANMMC_CAPABILITIES);

	ocr_avail = 0;
	if (caps & TITANMMC_CAN_VDD_330)
		ocr_avail |= MMC_VDD_32_33|MMC_VDD_33_34;
	if (caps & TITANMMC_CAN_VDD_300)
		ocr_avail |= MMC_VDD_29_30|MMC_VDD_30_31;
	if (caps & TITANMMC_CAN_VDD_180)
		ocr_avail |= MMC_VDD_165_195;
}

static int mmc_go_idle(void)
{
	uint *resp = NULL;
	resp = mmc_cmd(MMC_CMD_RESET, 0, MMC_RSP_NONE);
	if(!resp)
		return -1;

	return 0;
}

static int mmc_cmd_with_data(ushort cmd, 
		ulong arg, uint response_flags,
		uchar *dst, ulong len)
{
	uint *dst_u32 = (uint *)dst;
	uint *resp;
	uint intmask;
	if (len == 0)
		return 0;

	writeb(0xE, TITANMMC_PHY_BASE + TITANMMC_TIMEOUT_CONTROL);
	writew(TITANMMC_MAKE_BLKSZ(0, len), TITANMMC_PHY_BASE + TITANMMC_BLOCK_SIZE);
	writew(1, TITANMMC_PHY_BASE +TITANMMC_BLOCK_COUNT);

	/* setup data */
	writew(TITANMMC_TRNS_READ, TITANMMC_PHY_BASE + TITANMMC_TRANSFER_MODE);
	/* the byte address or the block address */
	/* send read command */
	resp = mmc_cmd(cmd, arg, response_flags | MMC_RSP_DATA);
	if(!resp)
		return -1;

	intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK);
	if(!intmask)
		return -1;

	if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
				TITANMMC_INT_DATA_CRC|
				TITANMMC_INT_DATA_END_BIT)) {
		mmc_err("mmc_cmd_with_data: err intmask%08x\n", intmask);
		return -EIO;
	}

	while (len>0) {
		*(dst_u32++) = readl(TITANMMC_PHY_BASE + TITANMMC_BUFFER);

		if (len >= 4)
			len -= 4;
		else {
			len = 0;
			break;
		}
	}

	mmc_debug("waiting for data complete interrupt\n");
	intmask = mmc_wait_interrupt(TITANMMC_INT_DATA_MASK );
	if(!intmask)
		return -EIO;

	if (intmask & (TITANMMC_INT_DATA_TIMEOUT|
				TITANMMC_INT_DATA_CRC|
				TITANMMC_INT_DATA_END_BIT)) {
		mmc_debug("mmc_cmd_with_data: err intmask %08x\n", intmask);
		return -EIO;
	}

	return 0;
}

static int mmc_app_cmd(void)
{
	uint *resp;
	resp = mmc_cmd(MMC_CMD_APP, rca << 16, MMC_RSP_R1);
	if (resp == NULL)
		return -1;

	/* Check that card supported application commands */
	if (!(resp[0] & MMC_R1_APP_CMD))
		return -1;

	udelay(5000);

	return 0;
}

int mmc_app_send_scr(void)
{
	int err;
	err = mmc_app_cmd();
	if (err)
		return err;

	udelay(1000);
	mmc_debug("raw scr before command 0x%x 0x%x \n", raw_scr[0], raw_scr[1]);
	err =mmc_cmd_with_data(SD_APP_SEND_SCR, 0, 
			MMC_RSP_R1, (uchar*)raw_scr, 
			sizeof(raw_scr)); 
	if(err)
		return err;

	mmc_debug("raw scr 0x%x 0x%x \n", raw_scr[0], raw_scr[1]);
	raw_scr[0] = ntohl(raw_scr[0]);
	raw_scr[1] = ntohl(raw_scr[1]);

	return 0;
}

static int mmc_send_if_cond(uint ocr)
{
	static const u8 test_pattern = 0xAA;
	ulong arg;
	uint *resp = NULL;

	/*
	 * To support SD 2.0 cards, we must always invoke SD_SEND_IF_COND
	 * before SD_APP_OP_COND. This command will harmlessly fail for
	 * SD 1.0 cards.
	 */
	arg = ((ocr & 0xFF8000) != 0) << 8 | test_pattern;

	resp = mmc_cmd(SD_SEND_IF_COND, arg, MMC_RSP_R7);
	if (resp == NULL)
		return -1;

	/* SD 2.0 cards return the test pattern
	 * in the response*/
	if ((resp[0] & 0xFF) != test_pattern)
		return -1;

	return 0;
}

static int mmc_send_app_op_cond(u32 ocr, u32 *rocr)
{
	int i, j,err = 0;
	uint *resp = NULL;

	/* the outer loop is repeated till the
	 * card returns the status as busy */
	for (i = 100; i; i--) {

		/* try 4 times */
		for (j = 0;j <=3;j++) {
			err = mmc_app_cmd();
			if (err){
				mmc_debug("cmd55 failed\n");
				continue;
			}
			mmc_debug("sending acmd 41\n");
			resp = mmc_cmd(SD_APP_OP_COND,ocr, MMC_RSP_R3);
			if (resp)
			{
				mmc_debug("acmd41 success \n");
				err = 0;
				break;
			}
		}
		if (err)
			break;

		/* if busy bit is set then the card is ready 
		 * else we need to repeat sending the cmd1*/
		/* while querying ocr we need not wait
		 * for busy bit to set */
		if (resp[0] & MMC_CARD_BUSY || ocr == 0)
			break;

		err = -1;

		udelay(90000);
	}

	if (rocr)
		*rocr = resp[0];

	return err;
}

static int mmc_all_send_cid(u32 *cid)
{
	uint *resp = NULL;

	resp = mmc_cmd(MMC_CMD_ALL_SEND_CID, 0, MMC_RSP_R2);
	if (!resp)
		return -1;

	memcpy(cid, resp, sizeof(u32) * 4);

	return 0;
}

static int mmc_set_relative_addr(void)
{
	uint *resp = NULL;

	resp = mmc_cmd(MMC_CMD_SET_RCA, rca<<16, MMC_RSP_R6);
	if (!resp)
		return -1;

	return 0;
}

static int mmc_send_relative_addr(uint *rca)
{
	uint *resp = NULL;

	resp = mmc_cmd(MMC_CMD_SET_RCA, 0, MMC_RSP_R6);
	if (!resp)
		return -1;

	*rca = resp[0] >> 16;

	return 0;
}

static int mmc_send_csd(u32 *csd)
{
	uint * resp = NULL;

	resp = mmc_cmd(MMC_CMD_SEND_CSD, rca << 16, MMC_RSP_R2 );
	if (!resp)
		return -1;

	memcpy(csd, resp, sizeof(u32) * 4);

	return 0;
}

static int mmc_select_card(uint rca)
{
	uint * resp = NULL;

	resp = mmc_cmd(MMC_CMD_SELECT_CARD, rca << 16, 
			rca?MMC_RSP_R1:MMC_RSP_NONE);
	if (!resp)
		return -1;

	return 0;
}

int mmc_app_set_bus_width(int width)
{
	int err;
	ulong arg;
	uint * resp = NULL;

	switch (width) {
		case MMC_BUS_WIDTH_1:
			arg = SD_BUS_WIDTH_1;
			break;
		case MMC_BUS_WIDTH_4:
			arg = SD_BUS_WIDTH_4;
			break;
		default:
			return -1;
	}
	err = mmc_app_cmd();
	if (err){
		mmc_debug("cmd55 failed\n");
		return err;
	}
	resp = mmc_cmd(SD_APP_SET_BUS_WIDTH, arg, MMC_RSP_R1);
	if (!resp)
	{
		printf("could not set the bus width\n");
		return err;
	}
	return 0;
}
/*
 * Handle the detection and initialisation of a card.
 */
static int mmc_init_card_sd( u32 ocr)
{
	int err;

	/*
	 * Since we're changing the OCR value, we seem to
	 * need to tell some cards to go back to the idle
	 * state.  We wait 1ms to give cards time to
	 * respond.
	 */
	mmc_go_idle();

	/*
	 * If SD_SEND_IF_COND indicates an SD 2.0
	 * compliant card and we should set bit 30
	 * of the ocr to indicate that we can handle
	 * block-addressed SDHC cards.
	 */
	err = mmc_send_if_cond(ocr);
	if (!err)
		ocr |= 1 << 30;

	err = mmc_send_app_op_cond(ocr, NULL);
	if (err)
		goto err;

	/*
	 * Fetch CID from card.
	 */
	err = mmc_all_send_cid(raw_cid);
	if (err)
		goto err;

	/*
	 * Set card RCA.
	 */
	err = mmc_send_relative_addr(&rca);
	if (err){
		printf("Error in send RCA command\n");
		goto err;
	}

	/*
	 * Fetch CSD from card.
	 */
	err = mmc_send_csd(raw_csd);
	if (err){
		printf("Error in send csd command\n");
		goto err;
	}

	err = mmc_decode_csd_sd();
	if (err){
		printf("Error in decoding csd\n");
		goto err;
	}

	mmc_decode_cid_sd();

	err = mmc_select_card(rca);
	if (err)
		goto err;
	err = mmc_app_send_scr();
	if (err){
		printf("Error in SCR command\n");
		goto err;
	}

	err = mmc_decode_scr();
	if (err)
		goto err;
	mmc_debug("Bus width supported 0x%x\n", scr.bus_widths);
	if ((scr.bus_widths & SD_SCR_BUS_WIDTH_4)) {
		err = mmc_app_set_bus_width(MMC_BUS_WIDTH_4);
		if (err )
		{
			mmc_err("appset buswidth failed\n");
			goto err;
		}

		mmc_set_buswidth(MMC_BUS_WIDTH_4);
	}
	if(!highcapacity)
	{
		uint *resp;
		resp = mmc_cmd(MMC_CMD_SET_BLOCKLEN, MMC_BLOCK_SIZE, MMC_RSP_R1);
		if(!resp)
			return -1;
	}
	return 0;
err:
	return -1;
}
/*
 * Mask off any voltages we don't support and select
 * the lowest voltage
 */
int mmc_switch_cmd(unsigned char set, unsigned char index, unsigned char value)
{
	ulong arg;
	uint *resp;
	arg = (MMC_SWITCH_MODE_WRITE_BYTE << 24) |
		(index << 16) |
		(value << 8) |
		set;

	resp = mmc_cmd(MMC_CMD_SWITCH, arg, MMC_RSP_R1B);
	if (!resp)
		return -1;
	else
		return 0;
}

u32 mmc_select_voltage(u32 ocr)
{
	int bit;
	int vdd = 0;

	ocr &= ocr_avail;

	bit = ffs(ocr);
	if (bit) {
		bit -= 1;
		ocr &= 3 << bit;
		switch(1 << bit)
		{
			case MMC_VDD_165_195:
				vdd |= TITANMMC_POWER_180;
				break;
			case MMC_VDD_29_30:
			case MMC_VDD_30_31:
				vdd |= TITANMMC_POWER_300;
				break;
			case MMC_VDD_32_33:
			case MMC_VDD_33_34:
				vdd |= TITANMMC_POWER_330;
				break;
			default:
				mmc_err("@@@@@@@@@@ Unknown voltage range\n");
		}

		mmc_set_power(vdd);
	} else {
		ocr = 0;
	}

	return ocr;
}

int mmc_init_sd(u32 ocr)
{
	/*
	 * Sanity check the voltages that the card claims to
	 * support.
	 */
	if (ocr & 0x7F) {
		mmc_err("TITANMMC: card claims to support voltages "			
				"below the defined range. These will be ignored.\n");
		ocr &= ~0x7F;
	}

	if (ocr & MMC_VDD_165_195) {
		mmc_err("TITANMMC: SD card claims to support the "
				"incompletely defined 'low voltage range'. This "
				"will be ignored.\n");
		ocr &= ~MMC_VDD_165_195;
	}

	ocr_selected = mmc_select_voltage(ocr);

	/*
	 * Can we support the voltage(s) of the card(s)?
	 */
	if (!ocr_selected)
		goto err;

	/*
	 * Detect and init the card.
	 */
	if(mmc_init_card_sd(ocr_selected))
		goto err;

	return 0;

err:
	return -1;
}
int mmc_send_op_cond(u32 ocr, u32 *rocr)
{
	int i, err = 0;
	uint *resp = NULL;

	for (i = 100; i; i--) {
		resp = mmc_cmd(MMC_CMD_SEND_OP_COND, ocr, MMC_RSP_R3);
		if (!resp){
			err = -1;
			break;
		}

		if (resp[0] & MMC_CARD_BUSY || ocr == 0){
			err = 0;
			break;
		}

		err = -1;
		udelay(10000);
	}

	if (rocr)
	{
		*rocr = resp[0];
		//printf("OCR rxd is = 0x%x\n", *rocr);
	}
	return err;
}


/*
 * Starting point for MMC card init.
 */
/*
 * Handle the detection and initialisation of a card.
 */
static int mmc_init_card_mmc(u32 ocr)
{
	int err;
	u32 ocr_return = 0;

	/*
	 * Since we're changing the OCR value, we seem to
	 * need to tell some cards to go back to the idle
	 * state.  We wait 1ms to give cards time to
	 * respond.
	 */
	mmc_go_idle();

	/* The extra bit indicates that we support high capacity */
	err = mmc_send_op_cond(ocr | (1 << 30), &ocr_return);
	//err = mmc_send_op_cond(ocr , NULL);
	if (err)
		goto err;

	if (ocr_return & 1 << 30)
	{
		printf("high capacity card\n");
		highcapacity = 1;
	}

	/*
	 * Fetch CID from card.
	 */
	err = mmc_all_send_cid(raw_cid);
	if (err)
		goto err;

	/*
	 * Set card RCA.
	 */
	err = mmc_set_relative_addr();
	if (err)
		goto err;

	/*
	 * Fetch CSD from card.
	 */
	err = mmc_send_csd(raw_csd);
	if (err)
		goto err;

	err = mmc_decode_csd_mmc();
	if (err)
		goto err;
	err = mmc_decode_cid_mmc();
	if (err)
		goto err;

	/*
	 * Select card, as all following commands rely on that.
	 */
	err = mmc_select_card(rca);
	if (err)
		goto err;

	if(csd.mmca_vsn >= CSD_SPEC_VER_4)
	{
		mmc_debug("setting bus width to 8\n");
		err = mmc_switch_cmd(EXT_CSD_CMD_SET_NORMAL,
				EXT_CSD_BUS_WIDTH, EXT_CSD_BUS_WIDTH_8);
		if(!err)
			mmc_set_buswidth(MMC_BUS_WIDTH_8);

	}
	if(!highcapacity)
	{
		uint *resp;
		resp = mmc_cmd(MMC_CMD_SET_BLOCKLEN, MMC_BLOCK_SIZE, MMC_RSP_R1);
		if(!resp)
			return -1;
	}
	return 0;

err:

	return -1;
}

int mmc_init_mmc(u32 ocr)
{
	int err;
	/*
	 * Sanity check the voltages that the card claims to
	 * support.
	 */
	if (ocr & 0x7F) {
		mmc_err("TITANMMC: card claims to support voltages "
				"below the defined range. These will be ignored.\n");
		ocr &= ~0x7F;
	}

	ocr_selected = mmc_select_voltage(ocr);

	/*
	 * Can we support the voltage of the card?
	 */
	if (!ocr_selected)
		goto err;

	/*
	 * Detect and init the card.
	 */
	err = mmc_init_card_mmc(ocr_selected);
	if (err)
		goto err;

	return 0;

err:
	return -1;
}

int mmc_init(int verbose)
{
	int retries, rc = -ENODEV;
	uint 	ocr;

	rca = 0;
	highcapacity = 0;
	highspeed = 0;
	mmc_ready = 0;
	/* reset the controller */
	mmc_reset(TITANMMC_RESET_ALL);

	/* enable interrupt status registers */
	mmc_enable_interrupt();

	/* set the clock for F_od (400K), set the bus width as 1 bit
	 * select the highest supported voltage */
	mmc_power_off();
	mmc_power_up();

	/* SDSpec requires waiting between power up and sending commands */
	udelay(5000);

	retries = 10;

	mmc_set_avail_ocr();

	mmc_debug("starting init\n");
	if(mmc_go_idle())
	{
		mmc_err("TITANMMC: Reset cmd failed\n");
		return rc;
	}

	/* this is required for SD2.0 cards; will fail quietly in 
	 * SD1.0 and MMC cards*/
	mmc_debug("sending cmd 8 for sd2.0 support\n");
	mmc_send_if_cond(ocr_avail);

	if (mmc_send_app_op_cond(0, &ocr) == 0) {
		mmc_debug("starting sd init\n");
		if (mmc_init_sd(ocr)){
			mmc_err("TITANMMC: No card found\n");
			mmc_power_off();
			return rc;
		}
	} else {
		/*
		 * If we fail to detect any SD cards then try
		 * searching for MMC cards.
		 */
		udelay(10000);
		if (mmc_send_op_cond(0, &ocr) == 0) {
			if(mmc_init_mmc(ocr))
			{
				mmc_err("TITANMMC: No card found\n");
				mmc_power_off();
				return rc;
			}
		}
		else {
			mmc_err("TITANMMC: No card found\n");
			mmc_power_off();
			return rc;
		}
	}
	if(csd.max_dtr > MAX_CARD_CLOCK_RATE)
		mmc_set_clock(MAX_CARD_CLOCK_RATE);
	else
	{
		mmc_set_clock(csd.max_dtr);
	}
	if(verbose)
	{
		print_cid();
	}
	mmc_dev.if_type=IF_TYPE_MMC;
	mmc_dev.part_type=PART_TYPE_DOS;
	mmc_dev.dev=0;
	mmc_dev.blksz=MMC_BLOCK_SIZE;
	sprintf(mmc_dev.vendor, "Man %02x %02x %02x Snr %02x %02x %02x", 0x00, 0x01, 0x02, 0x01, 0x02, 0x03);
	sprintf(mmc_dev.product, "%s", "MMC4");
	sprintf(mmc_dev.revision, "%x %x", 0x01, 0x01);
	mmc_dev.block_read=mmc_bread;

	mmc_ready = 1;
	return 0;
}

int mmc2info(ulong addr)
{
	/* FIXME hard codes to 32 MB device */
	if (addr >= CFG_MMC_BASE && addr < CFG_MMC_BASE + CFG_MMC_SIZE)
		return 1;

	return 0;
}

#endif	/* CONFIG_MMC */

