/*****************************************************************************
* Copyright 2006 - 2009 Broadcom Corporation.  All rights reserved.
*
* Unless you and Broadcom execute a separate written software license
* agreement governing use of this software, this software is licensed to you
* under the terms of the GNU General Public License version 2, available at
* http://www.broadcom.com/licenses/GPLv2.php (the "GPL"). 
*
* Notwithstanding the above, under no circumstances may you combine this
* software in any way with any other Broadcom software provided under a
* license other than the GPL, without Broadcom's express prior written
* consent.
*****************************************************************************/

/**
 * @file bcm59040_irqs.h
 * @brief Virtual IRQs for BCM59040 PMU
 *
 * This file is released under the GPLv2
 *
 */

#ifndef _BCM59040_IRQS_H_
#define _BCM59040_IRQS_H_

#define	IRQ_BCM59040_FIRST	NR_IRQS

enum BCM59040_IRQS
{
	/** @brief Power On Key/Button status interrupts
	 *
	 * The three events PONKEY Hold, PONKEY System Standby and PONKEY Restart
	 * all use different timers. The details can be found in the BCM59040 Data
	 * Sheet. PONKEY Hold is typically >1s and <5s. PONKEY System Standby is
	 * typically >5s and <15s. PONKEY System Standby is typically >15s.
	 *
	 * Included in this interrupt are:
	 *
	 *     - PONKEY Rising (after debounce time)
	 *     - PONKEY Falling (after debounce time)
	 *     - PONKEY Hold (PONKEY held low for preset time
	 *     - PONKEY System Standby (PONKEY held low for preset time
	 *     - PONKEY Restart (PONKEY held low for preset time
	 *     - PONKEY Restart On. Triggers after restart has completed
	 *       HOSTON is entered again.
	 */
	BCM59040_IRQ_PONKEY=IRQ_BCM59040_FIRST,

	BCM59040_IRQ_HIB_MODE,		/// @brief Hibernation mode
	BCM59040_IRQ_PMU_TOO_WARM,	/// @brief PMU is too warm (over temperature)

	/** @brief Wall Charger status interrupts
	 *
	 * Included in this interrupt are:
	 *
	 *     - Wall Charger Insertion
	 *     - Wall Charger Removal
	 *     - Wall Charger Over-voltage
	 */
	BCM59040_IRQ_WALL_CHARGER,

	/** @brief USB Charger status interrupts
	 *
	 * Included in this interrupt are:
	 *
	 *     - USB Charger Insertion
	 *     - USB Charger Removal
	 *     - USB Charger Over-voltage
	 */
	BCM59040_IRQ_USB_CHARGER,

	BCM59040_IRQ_EOC,			/// @brief End of charge interrupt, i.e., charge is completed
	BCM59040_IRQ_CHGDET_DONE,	/// @brief Charger detection is completed
	BCM59040_IRQ_VSR_OVERV,		/// @brief VSR is over-voltage
	BCM59040_IRQ_VSR_OVERC,		/// @brief VSR is over current limit

	/** @brief Charger error interrupts
	 *
	 * Included in this interrupt are:
	 *
	 *     - Charger not ok (shorted to ground)
	 *     - Charger watchdog alarm
	 *     - VBUS voltage collapse
	 *     - Charging error disappearance
	 *     - Charger watchdog alarm expiration
	 *     - Ideal diode overcurrent
	 */
	BCM59040_IRQ_CHRG_ERR,

	BCM59040_IRQ_REG_OVERC,		/// @brief Over-current indication for LDOs and regulators
								/// Readable virtual register provides bit mask of which
								/// LDOs or regulators are in overcurrent state.

	BCM59040_IRQ_REG_OVERV,		/// @brief Regulators in over-voltage condition
								/// Readable virtual register provides bit mask of which
								/// regulators are in over-voltage state.

	BCM59040_IRQ_BUP_BAT_LOW,	/// @brief Backup battery is low
	BCM59040_IRQ_FGC_COMPARE,	/// @brief FGC comparator interrupt

	BCM59040_IRQ_RTC_INVALID,	/// @brief RTC time is invalid, i.e., power lost
	BCM59040_IRQ_RTC_1S,		/// @brief RTC 1 second interrupt
	BCM59040_IRQ_RTC_60S,		/// @brief RTC 60 second interrupt
	BCM59040_IRQ_RTC_A1,		/// @brief RTC alarm 1 triggered


	/** @brief Main battery status interrupts
	 *
	 * Included in this interrupt are:
	 *
	 *     - temperature fault
	 *     - temperature too low
	 *     - temperature too high
	 *     - battery removed
	 *     - battery inserted
	 *     - battery low
	 *     - battery very low
	 */
	BCM59040_IRQ_MAIN_BATT,

/**@brief USB bus status changes
 *
 * Interrupts to be considered for this IRQ:
 *
 *     - VBUS valid comparator falling edge
 *     - Device A session valid falling edge
 *     - Device B session ended falling edge
 *     - ID insertion (IDOUT high to low transition)
 *     - VBUS valid comparator rising edge
 *     - Device A session valid rising edge
 *     - Device B session ended rising edge
 *     - ID removal (IDOUT low to high transition)
 *     - resume VBUS
 *     - IDIN value change
 *
 * @todo Need to better define this. There are multiple USB status interrupts
 * in BCM59040. I don't know whether this should be a single interrupt or
 * individual ones. It depends on whether it makes sense for a single ISR
 * to handle all events or if separate ISRs are likely for each event. Discuss
 * with Jim and Zhen.
 *
 */
	BCM59040_IRQ_USB,

/** @brief ADC completion interrupts
 *
 *     ADC conversion completed
 *     ADC data end
 *     ADC continuous conversion failed
 *     ADC asynchronous conversion off
 *     ADC asynchronous conversion failed
 */
	BCM59040_IRQ_SAR,

	BCM59040_FINAL_IRQ
}

/*
 * Redefine NR_IRQS since we just added quite a few.
 */

#undef NR_IRQS
#define NR_IRQS BCM59040_FINAL_IRQ

#endif

