/* 
 * Support for basic memory protection - enough to get the ARM DCache to
 * work, setting up a 1:1 mapping of the address space, on 1MB-grained
 * level
 *
 *  Copyright (C) 2009 TomTom
 *  Author: Martin Jackson <martin.jackson@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <common.h>
#include <asm/sizes.h>
#include <asm/memprot.h>

#define TRUNC_1MB(x) (x & 0xfff00000)
#define SHIFT_1MB(x) (x >> 20)
#define TTBL_LENGTH SZ_16K

static struct mem_area default_map[] = {
#ifdef PHYS_SDRAM_1
       {
               .base = PHYS_SDRAM_1,
               .length = PHYS_SDRAM_1_SIZE,
               .flags = MEMPROT_CACHED | MEMPROT_BUFFERED
       },
#endif
#ifdef PHYS_SDRAM_2
       {
               .base = PHYS_SDRAM_2,
               .length = PHYS_SDRAM_2_SIZE,
               .flags = MEMPROT_CACHED | MEMPROT_BUFFERED
       },
#endif
#ifdef PHYS_SDRAM_3
       {
               .base = PHYS_SDRAM_3,
               .length = PHYS_SDRAM_3_SIZE,
               .flags = MEMPROT_CACHED | MEMPROT_BUFFERED
       },
#endif
#ifdef PHYS_SDRAM_4
       {
               .base = PHYS_SDRAM_4,
               .length = PHYS_SDRAM_4_SIZE,
               .flags = MEMPROT_CACHED | MEMPROT_BUFFERED
       },
#endif
};

/* Section descriptor format:
 *
 * AAAAAAAAAAAA00000000PP0DDDD1CBSS
 * |          |--N/A---||-|  |-||^^---- 10 = section (1MB) descriptor
 * |          |        || |  | |^------ 1 = buffered, 0 = unbuffered
 * |          |        || |  | ^------- 1 = cached, 0 = uncached
 * |          |        || ^^^^--------- Domain control bits
 * |          |        ^^-------------- Access permission bits, set to 00
 * |          |                          to disallow access, 11 otherwise
 * ^^^^^^^^^^^^------------------------ Upper 12 bits of physical address
 *
 */
#define MD_ONES		0x00000010
#define MD_SECTION	0x00000002
#define MD_BUFFERED	0x00000004
#define MD_CACHED	0x00000008
#define MD_DOMAIN	0x00000000
#define MD_ALLOW	0x00000c00
#define MD_DEFAULT	MD_ONES | MD_ALLOW | MD_DOMAIN | MD_SECTION

uint32_t encode(struct mem_area *a)
{
#define GET_FLAG(f,y,n) (a->flags & f ? y : n)
	return MD_ONES |
		TRUNC_1MB((uint32_t) a->base) |
		GET_FLAG(MEMPROT_NOACCESS, 0x0, MD_ALLOW) |
		MD_DOMAIN |
		GET_FLAG(MEMPROT_CACHED, MD_CACHED, 0x0) |
		GET_FLAG(MEMPROT_BUFFERED, MD_BUFFERED, 0x0) |
		MD_SECTION;
}

#define set_ttb(x) __asm__ ( \
	"mcr p15, 0, %0, c2, c0, 0" ::"r"(x) \
)

#define set_access(x) __asm__ ( \
	"mcr p15, 0, %0, c3, c0, 0" ::"r"(x) \
)

#define flush_tlbs() { \
	int _t = 0; \
__asm__ ( \
	"mcr p15, 0, %0, c8, c7, 0" ::"r"(_t) /* Flush TLB */ \
); \
}

#define drain_wb() { \
	int _t = 0; \
__asm__ ( \
	"mcr p15, 0, %0, c7, c10, 4" ::"r"(_t) \
); \
}

/* TODO: Set alignment errors etc., Create error handler */
#define enable_mmu() \
{ \
	int _t=0; \
__asm__ __volatile__ ( \
	"mrc p15, 0, %0, c1, c0, 0 \n\t"	/* Read MMU ctl */ \
	"orr %0, %0, #0x1 \n\t" \
	"bic %0, %0, #0xc \n\t"		/* Disable Dcache, write buf */ \
	"b 1f \n\t" \
	/* --- Next I-cache line --- */ \
	".align 5 \n\t" \
	"1: mcr p15, 0, %0, c7, c7, 0\n\t"	/* Invalidate Caches */ \
	"mcr p15, 0, %0, c1, c0, 0 \n\t"	/* Enable MMU */ \
	"mrc p15, 0, %0, c1, c0, 0 \n\t"	/* Read back MMU ctl */ \
	/* --- Next I-cache line --- */ \
	".align 5 \n\t" \
	"nop" \
	: "+r"(_t) :: "r0" \
); \
}

/* Turn the mem_map table into a page table, enable the MMU */
void set_simple_map(uint32_t base, struct mem_area *mem_map, int map_len)
{

	int i=0;

	do {
#define MD_INIT(x) ((volatile uint32_t *) base)[i+x] = \
		MD_DEFAULT | (i+x)<<20

		MD_INIT(0x0);
skip_vectors:	MD_INIT(0x1); MD_INIT(0x2); MD_INIT(0x3);
		MD_INIT(0x4); MD_INIT(0x5); MD_INIT(0x6); MD_INIT(0x7);
		MD_INIT(0x8); MD_INIT(0x9); MD_INIT(0xa); MD_INIT(0xb);
		MD_INIT(0xc); MD_INIT(0xd); MD_INIT(0xe); MD_INIT(0xf);
		i+=16;
	} while(i<TTBL_LENGTH>>2);

#define MM_I mem_map[i]
	for(i=0; i<map_len; i++) {
		uint32_t j, md = encode(&MM_I);
		uint32_t l=SHIFT_1MB(MM_I.length), b = SHIFT_1MB(MM_I.base);

#define W ((uint32_t *) base)[j+b]
		for(j = 0; j < l; j++) {
			W = TRUNC_1MB(W) | md;
		}
	}

	set_ttb(base);
	set_access(0xffffffff); /* Set up all domains as 'Manager' */
	drain_wb();
	flush_tlbs();
	enable_mmu();
	flush_tlbs();
}

void set_default_map(uint32_t base)
{
	set_simple_map(base, default_map, ARRAY_SIZE(default_map));
}

