/*
 * U-Boot control program for TomTom devices
 *
 * Copyright (C) 2010 TomTom International B.V.
 * Author: Martin Jackson <martin.jackson@tomtom.com>
 * 
 ************************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 ************************************************************************
 */

/* #define DEBUG */

#include <alloca.h>
#include <assert.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <tt_header.h>

#define ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))
#define MAXPATH 256
#define SECTOR_SIZE 512

#define stringify(s)	tostring(s)
#define tostring(s)	#s

#ifdef VER
# define VERSION __DATE__ " " __TIME__ " " stringify(VER)
#else
# define VERSION __DATE__ " " __TIME__
#endif

static char uboot_partition[] = "/dev/uboot_part";

typedef int (*action_handler_t)(int argc, char *argv[]);

static int action_dump(int argc, char *argv[]);
static int action_read(int argc, char *argv[]);
static int action_write(int argc, char *argv[]);
static int action_upgrade(int argc, char *argv[]);

static struct {
	char *name;

#define BINFMT_FLAG_SIGNED 0x1
#define BINFMT_IGNORE_DEV  0x2	/* Whether to ignore /dev/uboot_part ? */
	int flags;
	size_t hdr_offset;
	int offset;	/* +ve is sector offset from start of volume,
			   -ve is sector offset from end */
}
binfmts[] = {
	{
#define BINFMT_BINARY 0 /* u-boot.bin */
		.name = "binary",
		.flags = 0x0,
		.hdr_offset = TT_HEADER_OFFSET,
	},
	{
#define BINFMT_BCM4760 1 /* u-boot.bcm4760 */
		.name = "bcm4760",
		.flags = BINFMT_FLAG_SIGNED,
		.hdr_offset = 440 + TT_HEADER_OFFSET,
	},
	{
#define BINFMT_S3CMMC 2 /* u-boot.s3cmmc */
		.name = "s3cmmc",
		.flags = BINFMT_FLAG_SIGNED | BINFMT_IGNORE_DEV,
		.hdr_offset = SECTOR_SIZE*1056 + TT_HEADER_OFFSET,
	},
};
#define get_binfmt(i) (binfmts[i])

static char progname[MAXPATH];

static struct {
	char *name;
	char *help;
	action_handler_t action;
}
actions[] = {
	{
		.name = "dump",
		.help = "dump [ -i <file> ]",
		.action = action_dump,
	},
	{
		.name = "read",
		.help = "read [ -i <file> ]",
		.action = action_read,
	},
	{
		.name = "write",
		.help = "write -v <version> -p platform [ -o <file> ] "
			"[ -b <bricknum> ] ",
		.action = action_write,
	},
	{
		.name = "upgrade",
		.help = "upgrade [-f] -i <file> [ -o <file> ]",
		.action = action_upgrade,
	},
};
#define A_I actions[i]

static struct {
	char *cpuinfo_name;
	char *plat_name;
} platforms[] = {

#define PLAT(x,y) {		\
	.cpuinfo_name = #x,	\
	.plat_name = #y,	\
}
#define BCM476X_PLAT(x) PLAT(x, bcm4760)

	BCM476X_PLAT(arezzo),
	BCM476X_PLAT(carlisle),
	BCM476X_PLAT(foggia),
	BCM476X_PLAT(lucca),
	BCM476X_PLAT(otavalo),
	BCM476X_PLAT(siena),
	PLAT(beaumont_11, beaumont_11),
	PLAT(havana, havana),
	PLAT(lima, lima),
	PLAT(torino_s, torinos), /* Watch out for the underscore */
	PLAT(venice, venice),
};

static void usage(void)
{
	int i;

	printf("USAGE: %s <action> ...\n", basename(progname));
	printf("USAGE: %s -v\n", basename(progname));
	puts("Where \'action\' is one of:\n");
	for(i=0; i<ARRAY_SIZE(actions); i++) {
		printf("\t%s\n", A_I.help);
	}
	printf("\nThe default file where uboot is located: %s\n",
	       uboot_partition);
	printf("%s header format: %d.%d\n", basename(progname),
	       TT_HEADER_MAJOR, TT_HEADER_MINOR);
}

static int read_header(FILE *f, long off, struct tt_header *raw)
{
	int r;

	if (fseek(f, off, SEEK_SET) == -1) {
		return -1;
	}

	r = fread(raw, sizeof(struct tt_header), 1, f);
	if (r == 0)
		return -1;

	return 0;
}

static void write_header(FILE *f, long off, struct tt_header *raw)
{
	if (fseek(f, off, SEEK_SET) == -1) {
		err(1, "Could not seek to header");
	}

	if (fwrite(raw, sizeof(struct tt_header), 1, f) == 0)
		err(1, NULL);
}

static int inline verify_magic(struct tt_header *hdr)
{
	return !strncmp((char *)&hdr->magic, TT_HEADER_MAGIC, 4);
}

/* Return index into binfmts, exit on error */
static int get_header(char *fname, struct tt_header *raw)
{
	int r = -1;
	FILE *f = fopen(fname, "rb");
	struct tt_header hdr;

	if (f == NULL) {
		err(1, "Could not open file %s", fname);
	}

	printf("Examining file [%s]...\n", fname);

	/* Try bcm4760 format */
	if (!read_header(f, get_binfmt(BINFMT_BCM4760).hdr_offset, raw))
	{
		__read_tt_header(raw, &hdr);
		if (verify_magic(&hdr)) {
			/* Could peek inside the BCM 440 byte header */
			r = BINFMT_BCM4760;
			goto end;
		}
	}

	/* Try s3cmmc format */
	if (!read_header(f, get_binfmt(BINFMT_S3CMMC).hdr_offset, raw))
	{
		__read_tt_header(raw, &hdr);
		if (verify_magic(&hdr)) {
			r = BINFMT_S3CMMC;
			goto end;
		}
	}

	/* Try boring binary format */
	if (!read_header(f, get_binfmt(BINFMT_BINARY).hdr_offset, raw))
	{
		__read_tt_header(raw, &hdr);
		if (verify_magic(&hdr)) {
			r = BINFMT_BINARY;
			goto end;
		}
	}

end:
	fclose(f);

	if (r < 0) {
		return -1;
	}

	printf("Binary type: \'%s\'\n", get_binfmt(r).name);
	return r;
	
}

static int get_header_fatal(char *fname, struct tt_header *raw)
{
	int r;

	if ((r = get_header(fname, raw)) < 0) {
		errx(1, "Unrecognized binary format");
	}

	return r;
}

/* Doesn't allocate anything */
static char *get_read_args(int argc, char *argv[])
{
	if (argc > 3) {
		goto err;
	} else if (argc == 3) {
		if (getopt(argc, argv, "i:") == -1) {
			goto err;
		}
		if (optarg == NULL)
			goto err;
		return optarg;
	} else { /* argc == 1 */
		return uboot_partition;
	}

err:
	fprintf(stderr, "Incorrect usage of \'%s\'\n", argv[0]);
	usage();
	exit(EXIT_FAILURE);
}

static int action_dump(int argc, char *argv[])
{
	struct tt_header raw;

	get_header_fatal(get_read_args(argc, argv), &raw);
	__dump_tt_header(&raw);

	return 0;
}

static int __check_version(unsigned maj, unsigned min, unsigned fatal)
{
#define MSG \
	"Binary has incompatible major version [%d] with this tool [%d]"

	if (maj != TT_HEADER_MAJOR) {
		if (fatal)
			errx(1, MSG, maj, TT_HEADER_MAJOR);
		warnx(MSG, maj, TT_HEADER_MAJOR);
		return -1;
	}
	if (min > TT_HEADER_MINOR) {
		warnx("Warning: Binary has greater minor version [%d] "
		      "than this tool (%d)", min, TT_HEADER_MINOR);
	}

	return 0;
}

#define check_version(x,y)		__check_version(x,y,0)
#define check_version_fatal(x,y)	__check_version(x,y,1)

static int action_read(int argc, char *argv[])
{
	struct tt_header raw, hdr;

	get_header_fatal(get_read_args(argc, argv), &raw);
	__read_tt_header(&raw, &hdr);

	check_version_fatal(hdr.ver_major, hdr.ver_minor);

	printf("TT Header version %d.%d\n", hdr.ver_major, hdr.ver_minor);
	printf("U-Boot changelist: %d\n", hdr.uboot_version);
	printf("U-Boot platform: %.*s\n", TT_HEADER_PLATLEN, hdr.platform);
	printf("U-Boot brick number: %d\n", hdr.brick);

	return 0;
}

#define case_opt_str(opt, out, e)			\
	case opt: if (optarg == NULL) goto e; 		\
		  out = optarg;				\
		  break;

#define case_opt_dec(opt, out, e)			\
	case opt: if (optarg == NULL) goto e;	 	\
		  errno = 0;				\
		  out = strtoul(optarg, NULL, 10);	\
		  assert(errno != EINVAL);		\
		  break;

#define case_opt_bool(opt, out, e)			\
	case opt: out = 1;				\
		  break;

/* Doesn't allocate anything */
static char *get_write_args(int argc, char *argv[], int *v, char **p,
			      int *b)
{
	int o;
	char *r = uboot_partition;

	*v = -1;
	*p = NULL;
	*b = TT_HEADER_BRICK_CURRENT;

	while ((o = getopt(argc, argv, "o:v:p:b:")) != -1) {
		switch (o) {
			case_opt_str('o', r, err);
			case_opt_dec('v', *v, err);
			case_opt_str('p', *p, err);
			case_opt_dec('b', *b, err);
			default:
				while(0);
		}
	}

	/* Mandatory args */
	if (*v == -1 || *p == NULL)
		goto err;

	return r;

err:
	fprintf(stderr, "Incorrect usage of \'%s\'\n", argv[0]);
	usage();
	exit(EXIT_FAILURE);
}

static int action_write(int argc, char *argv[])
{
	struct tt_header raw, hdr;
	int fd;
	FILE *f;
	int v,b;
	char *p, *fname;
	int binfmt;

	fname = get_write_args(argc, argv, &v, &p, &b);
	binfmt = get_header_fatal(fname, &raw);

	fd = open(fname, O_SYNC | O_WRONLY);
	if (fd == -1) {
		err(1, "Could not open file %s", fname);
	}
	f= fdopen(fd, "wb");
	if (f == NULL) {
		err(1, "Could not open file %s", fname);
	}

	if (get_binfmt(binfmt).flags & BINFMT_FLAG_SIGNED) {
		errx(1, "Writing header will invalidate signature, aborting");
	}
	memcpy((char *)&hdr.magic, TT_HEADER_MAGIC, sizeof(uint32_t));
	hdr.len = sizeof(struct tt_header);
	hdr.ver_major = TT_HEADER_MAJOR;
	hdr.ver_minor = TT_HEADER_MINOR;
	hdr.uboot_version = v;
	strncpy(hdr.platform, p, TT_HEADER_PLATLEN);
	hdr.brick = b;

	__write_tt_header(&hdr, &raw);
	write_header(f, get_binfmt(binfmt).hdr_offset, &raw);
	fclose(f);

	return 0;
}

/* Doesn't allocate anything */
static char *get_upgrade_args(int argc, char *argv[], char **d, int *f)
{
	int o;
	char *r = NULL;
	*f = 0;
	*d = uboot_partition;

	while ((o = getopt(argc, argv, "i:o:f")) != -1) {
		switch (o) {
			case_opt_str ('i', r, err);
			case_opt_str ('o', *d, err);
			case_opt_bool('f', *f, err);
			default:
				while(0);
		}
	}

	if (r != NULL)
		return r;

err:
	fprintf(stderr, "Incorrect usage of \'%s\'\n", argv[0]);
	usage();
	exit(EXIT_FAILURE);
}

static int verify_platform(char *plat)
{
	FILE *f;
	char platbuf[32];
	int i;
	char *plat1 = NULL, *r;

	errno = 0;
	/* Find hardware type from /proc/cpuinfo */
	f = popen("awk '/Hardware/ { printf \"%s\", tolower($3) }' /proc/cpuinfo",
		  "r");
	if (f == NULL) {
		err(1, "Could not detect hardware type");
	}
	r = fgets(platbuf, ARRAY_SIZE(platbuf), f);
	if (r == NULL) {
		warnx("Could not detect hardware type");
		return -1;
	}

	errno = 0;
	if (pclose(f) == -1)
		warn("Internal error: Could not close coprocess");

	for(i=0; i<ARRAY_SIZE(platforms); i++) {
		if (strcmp(platbuf, platforms[i].cpuinfo_name) == 0) {
			plat1 = platforms[i].plat_name;
			goto found_plat;
		}
	}

	warnx("Running on unrecognized platform");
	return -1;

found_plat:
	printf("Running on platform %s\n", plat1);

	if (strncmp(plat1, plat, TT_HEADER_PLATLEN)) {
		warnx("Bootloader does not match platform");
		return -1;
	}
	return 0;
}

static int verify_versions(uint32_t inver, uint32_t outver)
{
	if (inver > outver) {
		printf("Update file is newer (%d) than current (%d)\n",
		       inver, outver);
		return 0;
	}
	warnx("Update file is same/older (%d) than current (%d)", inver,
	      outver);
	return -1;
}

static int verify_bricknumber(uint32_t inbrick, uint32_t outbrick)
{
	if (inbrick >= outbrick) {
		printf("Update file has same/higher bricknumber (%d) "
		       "than current (%d)\n", inbrick, outbrick);
		return 0;
	}
	warnx("Update file has lower bricknumber (%d) than current (%d)",
	      inbrick, outbrick);
	return -1;
}

#define COPY_SIZE 1024*1024	/* 1MB */
static int write_update(char *infile, char *outfile)
{
	int in, out;
	void *buf;
	int r, w;
	int retries = 5;

start:
	if ((in = open(infile, O_RDONLY)) < 0)
		err(1, "Could not open input file");
	if ((out = open(outfile, O_WRONLY | O_SYNC)) < 0)
		err(1, "Could not open output file");

	if ((buf = mmap(NULL, COPY_SIZE, PROT_READ | PROT_WRITE,
			MAP_PRIVATE | MAP_ANONYMOUS, -1, 0))
	    == MAP_FAILED) {
		err(1, "Could not allocate mmap copy buffer");
	}

	while (1) {
		if ((r = read(in, buf, COPY_SIZE)) == -1) {
			warn("Could not read input file, retrying");
			goto retry;
		}
		if (r == 0)
			break;
		while (r > 0) {
			if ((w = write(out, buf, r)) == -1) {
				warn("Couldn't write output, retrying");
				goto retry;
			}
			r -= w;
		}
	}
	if (fsync(out) == -1) {
		if (errno == EINVAL)
			errx(1, "Output could not be synced");
		warn("Could not sync output file, retrying copy");
		goto retry;
	}

	printf("U-Boot update complete\n");

	r = close(in);
	if (r == -1)
		warn("Couldn't close input file");
	r = close(out);
	if (r == -1)
		warn("Couldn't close output file");

	return 0;

	/* For Broadcom, if it really goes wrong, we should erase stage II */

retry:
	r = close(in);
	if (r == -1)
		warn("Couldn't close input file");
	r = close(out);
	if (r == -1)
		warn("Couldn't close output file");
	if (--retries > 0)
		goto start;
	errx(1, "Writing u-boot failed");
}

static int action_upgrade(int argc, char *argv[])
{
	struct tt_header raw, hdr_in, hdr_out;
	int force, inhdr_valid = 1, outhdr_valid = 0;
	char *infile, *outfile;

	infile = get_upgrade_args(argc, argv, &outfile, &force),

	get_header_fatal(infile, &raw);
	__read_tt_header(&raw, &hdr_in);

	if (get_header(outfile, &raw) < 0) {
		warnx("Could not read output header");
	} else {
		__read_tt_header(&raw, &hdr_out);
		if (!check_version(hdr_out.ver_major, hdr_out.ver_minor))
			outhdr_valid = 1;
	}

	if (force) {
		if (check_version(hdr_in.ver_major, hdr_in.ver_minor))
			inhdr_valid = 0;
	} else {
		/* Refuse to upgrade if we don't understand the header */
		check_version_fatal(hdr_in.ver_major, hdr_in.ver_minor);
	}

	/* Do some sanity checks */
	if (inhdr_valid) {
#define FAIL_IF_NFORCE(x) { if((x) && !force) errx(1, "Aborting");}
		FAIL_IF_NFORCE(verify_platform(hdr_in.platform));
		if (outhdr_valid) {
			FAIL_IF_NFORCE(
				verify_versions(hdr_in.uboot_version,
						hdr_out.uboot_version));
			FAIL_IF_NFORCE(verify_bricknumber(hdr_in.brick,
				       hdr_out.brick));
		}
	}

	write_update(infile, outfile);

	return 0;
}

int main(int argc, char *argv[])
{
	int i;

	strncpy(progname, argv[0], MAXPATH-1);

	if (argc > 1 && !strcmp(argv[1],"-v")) {
		printf("Version: %s " VERSION "\n", basename(progname));
		exit(EXIT_SUCCESS);
	}

	if (argc < 2) {
		usage();
		exit(EXIT_FAILURE);
	}

	for(i=0; i<ARRAY_SIZE(actions); i++) {
		if (A_I.action != NULL && !strcmp(A_I.name, argv[1])) {
#ifdef DEBUG
			printf("Action = [%s]\n", A_I.name);
#endif
			exit(actions[i].action(argc-1, &argv[1]));
		}
	}

	fprintf(stderr, "Error: unrecognized option \"%s\"\n", argv[1]);
	usage();
	exit(EXIT_FAILURE);
}

